//
// Created by thejackimonster on 29.03.23. Modified by UnrealXR on 08.06.25.
//
// Copyright (c) 2023-2024 thejackimonster. All rights reserved.
// Copyright (c) 2025 UnrealXR. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "device_imu.h"
#include "device_mcu.h"

#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

enum XrealDriverCommand {
    ROLL,
    PITCH,
    YAW,
    GENERIC_MESSAGE,
    BRIGHTNESS_UP,
    BRIGHTNESS_DOWN,
};

int sock_fd;

void imu_event(uint64_t timestamp, device_imu_event_type event, const device_imu_ahrs_type* ahrs) {
	if (event != DEVICE_IMU_EVENT_UPDATE) {
		return;
	}

	device_imu_quat_type q = device_imu_get_orientation(ahrs);
	device_imu_euler_type e = device_imu_get_euler(q);

	char status_event = 0;
	uint32_t encoded_data;

	status_event = ROLL;
	memcpy(&encoded_data, &e.roll, sizeof(encoded_data));
	encoded_data = htonl(encoded_data);
    write(sock_fd, &status_event, 1);
    write(sock_fd, &encoded_data, sizeof(encoded_data));

    status_event = PITCH;
	memcpy(&encoded_data, &e.pitch, sizeof(encoded_data));
    encoded_data = htonl(encoded_data);
    write(sock_fd, &status_event, 1);
    write(sock_fd, &encoded_data, sizeof(encoded_data));

    status_event = YAW;
	memcpy(&encoded_data, &e.yaw, sizeof(encoded_data));
    encoded_data = htonl(encoded_data);
    write(sock_fd, &status_event, 1);
    write(sock_fd, &encoded_data, sizeof(encoded_data));
}

void mcu_event(uint64_t timestamp, device_mcu_event_type event, uint8_t brightness, const char* msg) {
    char status_event = 0;

	switch (event) {
		case DEVICE_MCU_EVENT_MESSAGE:
		    status_event = GENERIC_MESSAGE;
			write(sock_fd, &status_event, 1);

			uint32_t msg_len = strlen(msg);
			uint32_t len_net = htonl(msg_len);

			write(sock_fd, &len_net, sizeof(len_net));
			write(sock_fd, msg, msg_len);
			break;
		case DEVICE_MCU_EVENT_BRIGHTNESS_UP:
		    status_event = BRIGHTNESS_UP;
			write(sock_fd, &status_event, 1);
			write(sock_fd, &brightness, 1);
			break;
		case DEVICE_MCU_EVENT_BRIGHTNESS_DOWN:
		    status_event = BRIGHTNESS_DOWN;
		    write(sock_fd, &status_event, 1);
            write(sock_fd, &brightness, 1);
			break;
		default:
			break;
	}
}

int main(int argc, const char** argv) {
    printf("Connecting to Unix socket\n");
    char *unix_socket_path = getenv("UNREALXR_NREAL_DRIVER_SOCK");

    if (unix_socket_path == NULL) {
        printf("Unix socket not set!\n");
        return 1;
    }

    sock_fd = socket(AF_UNIX, SOCK_STREAM, 0);

    struct sockaddr_un server_addr;
    server_addr.sun_family = AF_UNIX;
    strcpy(server_addr.sun_path, unix_socket_path);

    int connection_result = connect(sock_fd, (struct sockaddr *)&server_addr, sizeof(server_addr));

    if (connection_result == -1) {
        perror("Failed to connect to Unix socket");
        exit(1);
    }

    printf("Connected to Unix socket\n");

	pid_t pid = fork();

	if (pid == -1) {
		perror("Could not fork!\n");
		return 1;
	}

	if (pid == 0) {
		device_imu_type dev_imu;
		if (DEVICE_IMU_ERROR_NO_ERROR != device_imu_open(&dev_imu, imu_event)) {
			return 1;
		}

		printf("Clearing and calibrating IMU\n");
		device_imu_clear(&dev_imu);
		device_imu_calibrate(&dev_imu, 1000, true, true, false);
		printf("Finished IMU tasks. Beginning data read\n");

		while (DEVICE_IMU_ERROR_NO_ERROR == device_imu_read(&dev_imu, -1));
		device_imu_close(&dev_imu);

		return 0;
	} else {
		int status = 0;

		device_mcu_type dev_mcu;
		if (DEVICE_MCU_ERROR_NO_ERROR != device_mcu_open(&dev_mcu, mcu_event)) {
			status = 1;
			goto exit;
		}

		printf("Reading MCU data\n");

		device_mcu_clear(&dev_mcu);
		while (DEVICE_MCU_ERROR_NO_ERROR == device_mcu_read(&dev_mcu, -1));
		device_mcu_close(&dev_mcu);

	exit:
		if (pid != waitpid(pid, &status, 0)) {
			return 1;
		}

		return status;
	}
}
